#![no_main]
#![no_std]

pub mod config;
pub mod multicast;

use alloc::boxed::Box;
use log::{debug, info};
use uefi::fs::PathBuf;
use uefi::prelude::*;
use uefi::proto::debug;
use uefi::proto::device_path::{DevicePath, LoadedImageDevicePath};
use uefi::table::boot::LoadImageSource;
use uefi::{Error, Result, Status};

use uefi::proto::device_path::text::{AllowShortcuts, DevicePathToText, DisplayOnly};
use uefi::proto::loaded_image::LoadedImage;
use uefi::table::boot::SearchType;
use uefi::Identify;
extern crate alloc;

fn load_efi_app_from_path(boot_services: &BootServices, app_path: Box<DevicePath>) -> Result {
    debug!("checking value for image path");
    let device_path_to_text_handle = *boot_services
        .locate_handle_buffer(SearchType::ByProtocol(&DevicePathToText::GUID))?
        .first()
        .expect("DevicePathToText is missing");

    let device_path_to_text =
        boot_services.open_protocol_exclusive::<DevicePathToText>(device_path_to_text_handle)?;

    let scoped_app_path = &*app_path;

    let image_device_path_text = device_path_to_text
        .convert_device_path_to_text(
            boot_services,
            scoped_app_path,
            DisplayOnly(true),
            AllowShortcuts(false),
        )
        .expect("convert_device_path_to_text failed");

    info!("Image path: {}", &*image_device_path_text);

    debug!("Loading image source");
    let image_source = LoadImageSource::FromDevicePath {
        device_path: scoped_app_path,
        from_boot_manager: false,
    };

    debug!("Getting image handler");
    let img_handler = boot_services.load_image(boot_services.image_handle(), image_source)?;

    debug!("Starting new image");
    boot_services.start_image(img_handler)
}

#[entry]
fn main(_handle: Handle, mut system_table: SystemTable<Boot>) -> Status {
    info!("init service");

    uefi_services::init(&mut system_table).unwrap();

    info!("init boot services");
    let boot_services = system_table.boot_services();

    info!("init runtime services");
    let runtime_services = system_table.runtime_services();

    let loaded_image = boot_services
        .open_protocol_exclusive::<LoadedImageDevicePath>(boot_services.image_handle())
        .unwrap();

    let device_path_to_text_handle = *boot_services
        .locate_handle_buffer(SearchType::ByProtocol(&DevicePathToText::GUID))
        .unwrap()
        .first()
        .expect("DevicePathToText is missing");

    let device_path_to_text = boot_services
        .open_protocol_exclusive::<DevicePathToText>(device_path_to_text_handle)
        .unwrap();

    let image_device_path_text = loaded_image
        .to_string(boot_services, DisplayOnly(false), AllowShortcuts(false))
        .expect("File path is not set")
        .expect("File path is not set 2");

    info!("Image path: {}", image_device_path_text);

    let conf = config::read_config(runtime_services).unwrap();
    let app_path = multicast::send_mcast(boot_services, &conf).unwrap();
    load_efi_app_from_path(boot_services, app_path).unwrap();

    Status::SUCCESS
}
